from argparse import ArgumentParser
import json
import os
import yaml

from ciou.types import ensure_list


def get_argparser():
    parser = ArgumentParser()
    parser.add_argument(
        'plan_file',
        type=str,
        nargs='*',
        help='Load requests plan from JSON or YAML file.')
    parser.add_argument(
        '--parallel',
        type=int,
        help='Limit number of parallel executions.')
    parser.add_argument(
        '-v', '--variable',
        action='append',
        dest='variables',
        metavar='NAME:VALUE',
        help=(
            'Define a variable to be used in the requests. '
            'To set multiple variables, use this argument multiple times'))
    parser.add_argument(
        '--no-animation',
        dest='animation',
        action='store_false',
        help='Disable progress animations in console output.')
    parser.add_argument(
        '--no-colors',
        dest='colors',
        action='store_false',
        help='Disable colors in console output.')
    parser.add_argument(
        '--version',
        action='store_true',
        help='Print version information.')

    return parser


def has_known_extension(path):
    for extension in ('.json', '.yaml', '.yml',):
        if path.endswith(extension):
            return True

    return False


def load_plan_files(paths, in_directory=False):
    plans = []

    paths = ensure_list(paths)
    if in_directory:
        paths = sorted(paths)

    for path in paths:
        if os.path.isdir(path):
            plans.extend(load_plan_files((os.path.join(path, i)
                         for i in os.listdir(path)), in_directory=True))
        elif has_known_extension(path) or not in_directory:
            plans.append(load_plan_file(path))

    return plans


def load_plan_file(filename):
    if not filename:
        raise ValueError('No input file given.')

    with open(filename, 'r') as f:
        if filename.endswith('.json'):
            plan = json.load(f)
        elif filename.endswith('.yaml') or filename.endswith('.yml'):
            plan = yaml.load(f, Loader=yaml.SafeLoader)
        else:
            raise ValueError(
                'Failed to recognize file type. '
                'File extension must be json, yaml, or yml.')

    plan['path'] = filename
    return plan


def parse_variables(raw_variables):
    variables = {}

    if not raw_variables:
        return variables

    for raw_variable in raw_variables:
        try:
            key, value = raw_variable.split(':', maxsplit=1)
            variables[key] = value
        except ValueError:
            raise ValueError(
                f'Variable definition "{raw_variable}" has invalid format. '
                'Variables should be defined as NAME:VALUE strings.')

    return variables



from datetime import datetime
from io import StringIO
from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool
from requests import request, Session
from requests.cookies import cookiejar_from_dict

from ciou.color import bold
from ciou.progress import MessageStatus, Update
from ciou.types import ensure_list

from .utils.template import Environment
from ._request import Request


PASS = 0
FAIL = 1
TOTAL = 2


class ListCounter:
    def __init__(self, input):
        if isinstance(input, list):
            self._data = list(input)
        else:
            self._data = [0] * input

    def increment(self, position):
        self._data[position] += 1

    def __getitem__(self, i):
        return self._data[i]

    @property
    def data(self):
        return list(self._data)

    def __add__(self, other):
        if isinstance(other, list):
            other = ListCounter(other)
        return ListCounter([sum(i) for i in zip(self.data, other.data)])


class PlansRunner:
    def __init__(self, plans, logger, parallel=None):
        self._plans = plans
        self._logger = logger
        self._parallel = min(len(self._plans),
                             parallel if parallel else cpu_count())

    def run(self):
        n_requests = ListCounter(3)
        n_plans = ListCounter(3)

        start = datetime.now()

        plans = ensure_list(self._plans)
        if self._parallel == 1:
            results = map(self._run_single_series, plans)
        else:
            self._logger.start()
            pool = ThreadPool(self._parallel)
            results = pool.map(self._run_single_parallel, self._plans)
            pool.close()
            self._logger.close()

        for n in results:
            n_requests += n
            n_plans.increment(FAIL if n[FAIL] else PASS)
            n_plans.increment(TOTAL)

        elapsed = (datetime.now() - start).total_seconds()

        summary = [
            ('Plans', n_plans.data),
            ('Requests', n_requests.data),
            ('Elapsed', f'{elapsed:.3f} s')
        ]
        if n_plans[TOTAL] == 1:
            summary = summary[1:]
        self._logger.summary(summary)

        return n_requests[FAIL]

    def _run_single_series(self, plan):
        display_filename = len(self._plans) > 1
        runner = PlanRunner(plan, self._logger, display_filename)
        return runner.run()

    def _run_single_parallel(self, plan):
        out = StringIO()
        logger = self._logger.copy(target=out, log_started=False)
        runner = PlanRunner(plan, logger, True, False)

        self._logger.push(Update(
            key=plan.path,
            message=bold(runner.title),
            status=MessageStatus.STARTED,
        ))

        n = runner.run()
        status = MessageStatus.ERROR if n[FAIL] else MessageStatus.SUCCESS
        out.seek(0)

        self._logger.push(Update(
            key=plan.path,
            message=bold(runner.title),
            details=out.read(),
            status=status,
        ))

        return n


class PlanRunner:
    def __init__(self, plan, logger, display_filename=False, print_name=True):
        self._plan = plan
        self._display_filename = display_filename
        self._print_name = print_name

        self._env = Environment()
        self._prepare_session()

        for name, value in self._env.resolve_templates(
                self._plan.variables).items():
            self._env.register(name, value)

        self._logger = logger

    def _prepare_session(self):
        session_option = self._plan.options.session
        if not session_option:
            self._session = None
            return

        self._session = Session()

        if isinstance(session_option, dict):
            session_dict = self._env.resolve_templates(session_option)
            self._session.headers.update(session_dict.get('headers', {}))
            self._session.cookies.update(
                cookiejar_from_dict(
                    session_dict.get(
                        'cookies', {})))

    def _request(self, *args, **kwargs):
        if self._plan.options.session:
            return self._session.request(*args, **kwargs)
        return request(*args, **kwargs)

    def _has_repeat_condition(self):
        return bool(self._plan.options.repeat_while)

    def _check_repeat_condition(self):
        repeat_while = self._plan.options.repeat_while
        if isinstance(repeat_while, str):
            return bool(self._env.resolve_expression(repeat_while))

        return bool(repeat_while)

    @property
    def title(self):
        name = self._plan.name
        if not self._display_filename:
            return name

        path = self._plan.path
        if name and path:
            return f'{name} ({path})'

        return path or name

    def run(self):
        n = ListCounter(3)

        repeat_index = 0 if self._has_repeat_condition() else None

        break_repeat = False
        repeat_while = True

        ignore_errors = self._plan.options.ignore_errors

        while repeat_while and not break_repeat:
            self._env.register('repeat_index', repeat_index)
            self._logger.title(
                self.title if self._print_name else None,
                len(self._plan.requests),
                repeat_index=repeat_index)

            self._logger.start()

            for request_dict in self._plan.requests:
                skip = not ignore_errors and n[FAIL] > 0

                request = Request(request_dict, self._env, skip)

                if request.state is None:
                    self._logger.start_request(request)
                    request.send(self._request)

                self._logger.finish_request(request)

                if not request.state.ok:
                    n.increment(FAIL)
                elif request.response is not None:
                    n.increment(PASS)
                n.increment(TOTAL)

            self._logger.close()

            break_repeat = not ignore_errors and n[FAIL] > 0
            repeat_while = (
                True if repeat_index == 0 else self._check_repeat_condition())

            if self._has_repeat_condition():
                repeat_index += 1

        return n.data



import json
import sys
import yaml

from ciou.color import bold, colors, fg_green, fg_red, fg_hi_black, no_color
from ciou.progress import Checks, MessageStatus, Progress, OutputConfig, Update

from .._request import RequestState


def get_assertion_status(assertion):
    if not assertion.executed:
        return MessageStatus.SKIPPED
    elif assertion.ok:
        return MessageStatus.SUCCESS
    else:
        return MessageStatus.ERROR


def get_status(state):
    if state in (RequestState.SUCCESS, RequestState.NOT_RAISED,):
        return MessageStatus.SUCCESS
    elif state in (RequestState.FAILURE, RequestState.ERROR,):
        return MessageStatus.ERROR
    elif state == RequestState.SKIPPED:
        return MessageStatus.SKIPPED


def _is_printable(pair):
    return isinstance(pair[1], (bool, dict, float, int, list, str))


class ConsoleLogger:
    def __init__(
            self,
            animations=True,
            colors=True,
            target=None,
            log_started=True):
        if not target:
            target = sys.stdout

        self._log_started = log_started

        self._output_config = OutputConfig(
            details_color=no_color,
            disable_animation=(not animations),
            disable_colors=(not colors),
            target=target,
        )
        self._progress = None

    def copy(self, **kwargs):
        current = dict(
            animations=(not self._output_config.disable_animation),
            colors=(not self._output_config.disable_colors),
            target=self._output_config.target,
        )
        return ConsoleLogger(**{**current, **kwargs})

    def start(self):
        self._progress = Progress(config=self._output_config)
        self._progress.start()

    def _style(self, text, *color):
        if self._output_config.disable_colors:
            return text

        return colors(*color)(text)

    def _print(self, *args):
        return print(*args, file=self._output_config.target)

    def close(self):
        if self._progress:
            self._progress.stop()

    def error(self, error):
        text = str(error)
        if not text:
            return

        error_text = self._style('ERROR:', bold, fg_red)
        self._print(f'{error_text} {text}')

    def _repeat_text(self, repeat_index):
        if repeat_index is None:
            return ''
        return f' (repeat_index={repeat_index})'

    def title(self, name, num_requests, repeat_index=None):
        name_text = f'{self._style(name, bold)}\n' if name else ''
        self._print(
            f'{name_text}Sending {num_requests} requests'
            f'{self._repeat_text(repeat_index)}:\n')

    def push(self, update):
        return self._progress.push(update)

    def summary(self, rows):
        key_width = max(len(i[0]) for i in rows) + 1

        for key, value in rows:
            if isinstance(value, list):
                passed, failed, total = value
                values = []

                if passed:
                    values.append(
                        self._style(
                            f'{passed} succeeded',
                            bold,
                            fg_green))
                if failed:
                    values.append(
                        self._style(
                            f'{failed} failed',
                            bold,
                            fg_red))

                values.append(f'{total} total')
                value = ', '.join(values)
            self._print(
                f'{self._style((key + ":").ljust(key_width), bold)} {value}')

    def _get_name_text(self, request):
        if request.name:
            return self._style(request.name, bold)
        return ''

    def _get_method_text(self, request):
        return (
            f'{self._style(request.method, bold)} {request.params.get("url")}')

    def _get_response_code_text(self, request):
        response = request.response
        if response is None:
            return ''

        code_text = self._style(f'HTTP {response.status_code}', bold)
        elapsed_ms = response.elapsed.total_seconds() * 1000
        not_raised_text = ''
        if request.state == RequestState.NOT_RAISED:
            not_raised_text = self._style(
                ' HTTP status code ignored.', fg_hi_black)

        return f'{code_text} ({elapsed_ms:.3f} ms){not_raised_text}'

    def _get_message_text(self, request):
        message = request.state.message

        type_text = self._style(f'{str(request.state).upper()}:', bold)
        return f'{type_text} {message}' if message else ''

    def _get_assertion_text(self, request):
        if request.state == RequestState.SKIPPED:
            return ''

        checks = Checks()
        for assertion in request.assertions:
            checks.push(Update(
                message=assertion.name,
                status=get_assertion_status(assertion),
            ))

        text = checks.getvalue()
        return f'\n{text}' if text else ''

    def _headers_text(self, headers):
        return '\n'.join(
            f'{self._style(key, bold)}: {value}' for key,
            value in headers.items())

    def _variables_text(self, variables):
        return '\n'.join(
            f'{self._style(key, bold)}: {value}' for key,
            value in dict(filter(_is_printable, variables.items())).items())

    def _body_text(self, body, content_type):
        if not isinstance(body, str):
            body = body.decode('utf-8')

        if content_type.startswith('application/json'):
            return json.dumps(
                json.loads(body), indent=2)
        elif content_type.startswith('application/yaml'):
            return yaml.dump(
                yaml.safe_load(body), default_flow_style=False)
        return body

    def _response_output_text(self, request, output):
        response = request.response

        def _format_output(output, prefix=''):
            output = output.rstrip(' \n').replace('\n', f'\n{prefix}')
            if not output.endswith('\n'):
                output = f'{output}\n'
            return f'\n{prefix}{output}'

        try:
            if not output:
                return ''
            elif (output.lower() == 'headers' or
                    output.lower() == 'response_headers'):
                return _format_output(
                    self._headers_text(response.headers), '< ')
            elif output.lower() == 'request_headers':
                return _format_output(
                    self._headers_text(response.request.headers), '> ')
            elif output.lower() == 'request_body':
                raw_body = response.request.body
                content_type = response.request.headers.get('Content-Type')
                return _format_output(
                    self._body_text(raw_body, content_type), '> ')
            elif output.lower() == 'response_body':
                content_type = response.headers.get('Content-Type')
                return _format_output(
                    self._body_text(response.text, content_type), '< ')
            elif output.lower() == 'text':
                return _format_output(response.text, '< ')
            elif output.lower() == 'json':
                pretty_json = json.dumps(response.json(), indent=2)
                return _format_output(pretty_json, '< ')
            elif output.lower() == 'variables':
                return _format_output(
                    self._variables_text(request._template_env.globals))
            elif output.lower() in ('yml', 'yaml'):
                pretty_yaml = yaml.dump(
                    response.json(), default_flow_style=False)
                return _format_output(pretty_yaml, '< ')
            else:
                return _format_output(
                    f'Unknown output entry [{output}], expected one of [\
headers, request_headers, request_body, \
response_headers, response_body, text, json, variables, \
yml, yaml]', '? ')
        except BaseException:
            return ''

    def _response_text(self, request):
        output = request.options.output
        output = output if isinstance(output, list) else [output]

        return ''.join(self._response_output_text(request, i) for i in output)

    def start_request(self, request):
        if not self._log_started:
            return

        text = self._get_name_text(request) or self._get_method_text(request)

        self._progress.push(Update(
            key=request.id,
            message=text,
            status=MessageStatus.STARTED,
        ))

    def finish_request(self, request):
        name_text = self._get_name_text(request)
        message = name_text or self._get_method_text(request)

        method_text = name_text and f'{self._get_method_text(request)}\n'
        code_text = self._get_response_code_text(request)
        message_text = self._get_message_text(request)
        message_separator = '\n  ' if message_text and code_text else ''

        details = (
            f'{method_text}'
            f'{code_text}{message_separator}{message_text}\n'
            f'{self._get_assertion_text(request)}'
            f'{self._response_text(request)}\n')

        self._progress.push(Update(
            key=request.id,
            message=message,
            details=details,
            status=get_status(request.state)
        ))
